
/*
 * Copyright (C) 2009 Tudor Marian (tudorm@cs.cornell.edu)
 */

#ifndef __MAEL_HDR__
#define __MAEL_HDR__

#include "xors.h"

/* yanked straight out of net/core/kmap_skb.h */
static inline void *kmap_skb_frag(const skb_frag_t *frag)
{
#ifdef CONFIG_HIGHMEM
	BUG_ON(in_irq());

	local_bh_disable();
#endif
	return kmap_atomic(frag->page, KM_SKB_DATA_SOFTIRQ);
}
static inline void kunmap_skb_frag(void *vaddr)
{
	kunmap_atomic(vaddr, KM_SKB_DATA_SOFTIRQ);
#ifdef CONFIG_HIGHMEM
	local_bh_enable();
#endif
}

#define memxor xor_bytes
/* this should not be inlined */
static inline int skb_xor_bits(struct sk_buff *skb, int offset, void *to, int len)
{
	int i, copy;
	int start = skb_headlen(skb);

	if (offset + len > (int) skb->len)
		goto fault;

	// copy the header
	if ((copy = start - offset) > 0)
	{
		if (copy > len)
			copy = len;
		memxor(to, skb->data + offset, copy);
		if ((len-= copy) == 0)
			return 0;
		offset += copy;
		to += copy;
	}

	for (i = 0; i < skb_shinfo(skb)->nr_frags; i++)
	{
		int end;

		WARN_ON(start > offset + len);

		end = start + skb_shinfo(skb)->frags[i].size;
		if ((copy = end - offset) > 0)
		{
			u8 *vaddr;

			if (copy > len)
				copy = len;

			vaddr = kmap_skb_frag(&skb_shinfo(skb)->frags[i]);
			memxor(to, vaddr + skb_shinfo(skb)->frags[i].page_offset +
					offset - start, copy);
			kunmap_skb_frag(vaddr);

			if ((len -= copy) == 0)
				return 0;
			offset += copy;
			to += copy;
		}
		start = end;
	}

	if (skb_shinfo(skb)->frag_list)
	{
		struct sk_buff *list = skb_shinfo(skb)->frag_list;

		for (; list; list = list->next)
		{
			int end;

			WARN_ON(start > offset + len);

			end = start + list->len;
			if ((copy = end - offset) > 0)
			{
				if (copy > len)
					copy = len;
				if (skb_xor_bits(list, offset - start, to, copy))
					goto fault;
				if ((len -= copy) == 0)
					return 0;
				offset += copy;
				to += copy;
			}
		}
	}

	if (!len)
		return 0;

	fault:
		return -EFAULT;
}

static int ksock_snd(struct socket *sock, struct sockaddr_in *addr, 
	void *buf, int len)
{
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
	int size;

	iov.iov_base = buf;
	iov.iov_len = len;

	msg.msg_flags = 0;
	if (addr == NULL)
	{
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
	}
	else
	{
		msg.msg_name = addr;
		msg.msg_namelen = sizeof(struct sockaddr_in);
	}
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	size = sock_sendmsg(sock,&msg,len);
	set_fs(oldfs);

	return size;
}

int ksock_snd_iov(struct socket *sock, struct sockaddr_in *addr, 
		struct iovec* iov, int iovcnt)
{
	struct msghdr msg;
	mm_segment_t oldfs;
	int size;

	msg.msg_flags = 0;
	if (addr == NULL)
	{
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
	}
	else
	{
		msg.msg_name = addr;
		msg.msg_namelen = sizeof(struct sockaddr_in);
	}
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov = iov;
	msg.msg_iovlen = iovcnt;
	msg.msg_control = NULL;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	size = sock_sendmsg(sock, &msg, iov_length(iov, iovcnt)); /* scatter */ 
	set_fs(oldfs);

	return size;
}

static int ksock_rcv(struct socket* sock, struct sockaddr_in* addr, 
		void* buf, int len)
{
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
	int size;

	iov.iov_base = buf;
	iov.iov_len = len;

	msg.msg_flags = 0;
	msg.msg_name = addr;
	msg.msg_namelen = sizeof(struct sockaddr_in);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	size = sock_recvmsg(sock,&msg,len,msg.msg_flags);
	set_fs(oldfs);

	return size;
}

/* sending an IP packet using a raw socket -- used to send recovered packets */
static inline int raw_send(struct socket *raw_sock, struct iphdr *packet)
{
	struct sockaddr_in raw_dest_addr;
	/* only works if the packet is an IP packet */
	int size;
	struct iphdr *iph = (struct iphdr *)(packet);
	int ip_len = ntohs(iph->tot_len); /* includes the length of the header */
	memset(&raw_dest_addr, 0, sizeof(struct sockaddr_in));
	raw_dest_addr.sin_family = AF_INET;
	raw_dest_addr.sin_addr.s_addr = iph->daddr;

	/* regular send with a null address */
	size = ksock_snd(raw_sock, &raw_dest_addr, packet, ip_len);
	return size;
}

#endif // __MAEL_HDR__
