/*
 * Copyright (C) 2009 Tudor Marian (tudorm@cs.cornell.edu)
 */

#ifndef _MAELSTROM_XORS_H
#define _MAELSTROM_XORS_H

#define SSE_CHUNK	(16)

#ifdef __KERNEL__
#include <linux/preempt.h>
#define SSE_START(cr0) do { \
	preempt_disable(); \
	cr0 = read_cr0(); \
	clts(); \
	} while(0)


#define SSE_END(cr0) do { \
	write_cr0(cr0); \
	preempt_enable(); \
	} while(0)
#endif //__KERNEL__

static inline void xor_bytes_bytewise(void *dest, void *src, int len)
{
	int i;
	for (i = 0; i < len; i++)
		((char *) dest)[i] ^= ((char *) src)[i];
}

/* 
 * xors length bytes from src starting at position 0 into dest starting at 
 * offset long is the size of the register, hence on x86_64 sizeof(long) == 8
 */
static inline void xor_bytes_registerwise(void *dest, void *src, int len)
{
	int i;

	for (i = 0; i < (len / sizeof(long)); i++)
		((long *) dest)[i] ^= ((long *) src)[i];

	// the remaining bytes
	for (i = (len / sizeof(long))*sizeof(long); i < len; i++)
		((char *) dest)[i] ^= ((char *) src)[i];
}

static inline void xor_bytes_sse2(void *dest, void *src, int len)
{
	int i;
#ifdef __KERNEL__
	unsigned char xmm0_save[16];
	unsigned char xmm1_save[16];
	unsigned int cr0;

	SSE_START(cr0);
	asm volatile
	(
			"movupd %%xmm0, (%0) \n\t"
			"movupd %%xmm1, (%1) \n\t"
			:
			: "r" (xmm0_save), "r" (xmm1_save)
	);
#endif

	for (i = 0; i < len / SSE_CHUNK; i++)
	{
		asm volatile
		(		// warning: cannot use memory src operand unless 16 byte aligned
				"movupd (%0), %%xmm0 \n\t"
				"movupd (%1), %%xmm1 \n\t"
				"xorpd %%xmm1, %%xmm0 \n\t"
				"movupd %%xmm0, (%0) \n\t"
				:
				: "r" (dest), "r" (src)
				: "memory"
		);

		dest += SSE_CHUNK;
		src += SSE_CHUNK;
	}

#ifdef __KERNEL__
	asm volatile
	(
			"sfence\n\t"
			"movupd (%0), %%xmm0\n\t"
			"movupd (%1), %%xmm1\n\t"
			:
			: "r" (xmm0_save), "r" (xmm1_save)
			: "memory"
	);
	SSE_END(cr0);
#endif

	xor_bytes_registerwise(dest, src, (len % SSE_CHUNK));
}

//#define xor_bytes	xor_bytes_sse2
#define xor_bytes	xor_bytes_registerwise

#endif //MAELSTROM

